require 'puppet/parameter/boolean'
require 'fileutils'
require 'net/http'
require 'uri'
require 'json'
require 'base64'
require 'openssl'

Puppet::Type.type(:cfssl).provide(:cfssl) do
    desc 'cfssl'

    def create
        # If authkey is absent
        #   The request is striaght forward. Build the request object 
        #   and submit to /api/v1/cfssl/newcert to receive back
        #   cert, key, csr. Save to the specified locations
        #   Done
        #   https://github.com/cloudflare/cfssl/blob/master/doc/api/endpoint_newcert.txt
        # else if we are performing a signed request we need to
        #   request the key and csr first and post the csr to an authsign endpoint
        #   https://github.com/cloudflare/cfssl/blob/master/doc/api/endpoint_newkey.txt
        #   - AND -
        #   https://github.com/cloudflare/cfssl/blob/master/doc/api/endpoint_authsign.txt
        # endif
        #
        # Finally we need to retrieve back the certificate chain. This is done by 
        #
        #response = Hash.new
        
        cfssl = Cfssl.new(resource[:server])
        result = Hash.new

        request_options = Hash.new
        request_options[:profile] = resource[:profile] if resource[:profile] 
        request_options[:label] = resource[:label] if resource[:label] 
        request_options[:bundle] = resource[:bundle] if resource[:bundle] 
        request_options[:subject] = resource[:subject] if resource[:subject] 
        request_options[:hosts] = resource[:hosts] if resource[:hosts] 
        request_options[:serial_sequence] = resource[:serial_sequence] if resource[:serial_sequence] 

        keydata = cfssl.newkey(resource[:manifest])
        result[:private_key] = keydata[:result][:private_key]
        result[:certificate_request] = keydata[:result][:certificate_request]

        if resource[:authkey]
            certdata = cfssl.authsign(resource[:authkey], result[:certificate_request], request_options)
            result[:certificate] = certdata[:result][:certificate]
        else
            certdata = cfssl.sign(result[:certificate_request], request_options)
            result[:certificate] = certdata[:result][:certificate]
        end

        File.write(certfile_crt, result['certificate'])
        File.write(certfile_key, result['private_key'])
        File.write(certfile_csr, result['certificate_request'])
    end

    def destroy
        FileUtils.rm(certfile_crt) if File.exist?(certfile_crt)
        FileUtils.rm(certfile_key) if File.exist?(certfile_key)
        FileUtils.rm(certfile_csr) if File.exist?(certfile_csr)

        # We should also here revoke the certificate, maby in version 2.
    end

    def exists?
        # we need to check that
        #   A) The certificate has not been revoked
        #   B) The certificate is not close to expiry. DONE
        #   C) The certificate matches the requested information
        response = false
        if File.exist?(certfile_crt)
            cert_data = File.read(certfile_crt)
            begin
                cert = OpenSSL::X509::Certificate.new(cert_data)
            rescue
                # Invalid Certificate
                return false
            end
            if ((cert.not_after.to_i - Time.now.to_i) < (60*60*24*30))
                # if the certificate expires within 30 days 
                # we return false instantly to regenerate
                return false
            end
            response = true
        end
        response = true if File.exist?(certfile_key)
        response = true if File.exist?(certfile_csr)

        response        
    end

private

    class Cfssl
        def initialize(host)
            host =~ /^https/ ? ssl = true : ssl = false
            hosturi = URI(host)
            conn = Net::HTTP.new(hosturi.host, hosturi.port)
            conn.use_ssl = ssl
            @conn = conn
        end

        def authsign(authkey, csr, options = {})
            inner_request = {
               :certificate_request => csr 
            }.merge(options)

            inner_request_json = JSON.generate(inner_request)
            token = OpenSSL::HMAC.digest("SHA256", hexdecode(authkey), inner_request_json)

            outer_request = {
                :token => Base64.strict_encode64(token),
                :request => Base64.strict_encode64(inner_request_json)
            }
            response = @conn.post('/api/v1/cfssl/authsign', JSON.generate(outer_request))
            process_response(response)
        end

        def sign(csr, options = {})
            request = {
               :certificate_request => csr 
            }.merge(options)

            response = @conn.post('/api/v1/cfssl/sign', JSON.generate(request))
            process_response(response)
        end

        def newkey(manifest)
            response = @conn.post('/api/v1/cfssl/newkey', JSON.generate(manifest))
            process_response(response)
        end

        def newcert(manifest, options)
            request = {
               :request => manifest 
            }.merge(options)

            response = @conn.post('/api/v1/cfssl/newcert', JSON.generate(request))
            process_response(response)
        end

        private

        def process_response(response)
            begin
                body = JSON.parse(response.body, {:symbolize_names => true})
            rescue => e
                raise e
            end

            if !body[:success]
                raise body[:errors][0][:message]
            end
                
            body
        end
        
        def hexdecode(s)
            s.scan(/../).map { |x| x.hex }.pack('c*')
        end 
    end

    def certfile_crt
        "#{certbase}.crt"
    end

    def certfile_key
        "#{certbase}.key"
    end

    def certfile_csr
        "#{certbase}.csr"
    end

    def certfile_chain
        "#{certbase}_chain.crt"
    end

    def certbase
        "#{resource[:certpath]}/cfssl_#{resource[:name]}"
    end
end
